/*
 * $Id: choices.c 5172 2007-12-05 20:46:35Z realh $
 *
 * ROX-Session, a very simple session manager
 * Copyright (C) 2002, the ROX-Filer team.
 *
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the Free
 * Software Foundation; either version 2 of the License, or (at your option)
 * any later version.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
 * more details.
 *
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, write to the Free Software Foundation, Inc., 59 Temple
 * Place, Suite 330, Boston, MA  02111-1307  USA
 */
/* choices.c - code for handling loading and saving of user choices */

#include "config.h"

#include <stdio.h>
#include <stdlib.h>
#include <sys/param.h>
#include <fcntl.h>
#include <errno.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <unistd.h>

#include "global.h"

#include "choices.h"

static gchar **dir_list = NULL;

/* Static prototypes */
inline static gboolean exists(const char *path)
{
	return g_file_test(path, G_FILE_TEST_EXISTS);
}


/****************************************************************
 *			EXTERNAL INTERFACE			*
 ****************************************************************/


/* Reads in XDG_CONFIG_* and constructs the directory list table.
 * You must call this before using any other choices_* functions.
 *
 * If XDG_CONFIG_* do not exist then suitable defaults are used.
 */
void choices_init(void)
{
#define ROX_SF_NET "rox.sourceforge.net"
	const gchar * const *e_dirs;
	int n;

	g_return_if_fail(dir_list == NULL);

	e_dirs = g_get_system_config_dirs();
	for (n = 0; e_dirs[n]; ++n);
	dir_list = g_new(gchar *, n + 2);
	dir_list[0] = g_build_filename(g_get_user_config_dir(), ROX_SF_NET, NULL);
	for (n = 0; e_dirs[n]; ++n)
		dir_list[n + 1] = g_build_filename(e_dirs[n], ROX_SF_NET, NULL);
	dir_list[n + 1] = NULL;
}

/* Returns an array of the directories in <Choices> which contain
 * a subdirectory called 'dir'.
 *
 * Lower-indexed results should override higher-indexed ones.
 *
 * Free the list using choices_free_list().
 */
GPtrArray *choices_list_dirs(char *dir)
{
	GPtrArray	*list;
	gchar		**cdir = dir_list;

	g_return_val_if_fail(dir_list != NULL, NULL);

	list = g_ptr_array_new();

	while (*cdir)
	{
		guchar	*path;

		path = g_strconcat(*cdir, "/", dir, NULL);
		if (exists(path))
			g_ptr_array_add(list, path);
		else
			g_free(path);

		cdir++;
	}

	return list;
}

void choices_free_list(GPtrArray *list)
{
	guint	i;

	g_return_if_fail(list != NULL);

	for (i = 0; i < list->len; i++)
		g_free(g_ptr_array_index(list, i));

	g_ptr_array_free(list, TRUE);
}

/* Get the pathname of a choices file to load. Eg:
 *
 * choices_find_path_load("menus", "ROX-Filer")
 *		 		-> "/etc/xdg/rox.sourceforge.net/ROX-Filer/menus".
 *
 * The return values may be NULL - use built-in defaults.
 * g_free() the result.
 */
guchar *choices_find_path_load(char *leaf, char *dir)
{
	gchar	**cdir = dir_list;

	g_return_val_if_fail(dir_list != NULL, NULL);

	while (*cdir)
	{
		gchar	*path;

		path = g_strconcat(*cdir, "/", dir, "/", leaf, NULL);

		if (exists(path))
			return path;

		g_free(path);

		cdir++;
	}

	return NULL;
}

/* Returns the pathname of a file to save to. If 'create' is TRUE then
 * intermediate directories will be created (set this to FALSE if you just
 * want to find out where * a saved file would go without actually altering
 * the filesystem).
 *
 * g_free() the result.
 */
guchar *choices_find_path_save(char *leaf, char *dir, gboolean create)
{
	gchar	*path, *retval;
	
	g_return_val_if_fail(dir_list != NULL, NULL);

	path = g_strconcat(dir_list[0], "/", dir, NULL);
	if (create && !exists(path))
	{
		if (g_mkdir_with_parents(path, 0777))
		{
			g_warning("g_mkdir_with_parents(%s): %s\n",
					path, g_strerror(errno));
		}
	}

	retval = g_strconcat(path, "/", leaf, NULL);
	g_free(path);

	return retval;
}

